import argparse

import functools
import math
import os.path
import sys
from datetime import datetime
from typing import Mapping, Any

import PIL.Image
import netron
import torch
import numpy as np
import matplotlib.pyplot as plt
import torchvision
import torchvision.transforms.functional as F
from torch import nn
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
from torchvision.utils import make_grid, draw_bounding_boxes
from torchvision.io import read_image
from pathlib import Path
from torchvision.models.detection import maskrcnn_resnet50_fpn, MaskRCNN_ResNet50_FPN_Weights
from torchvision.models.detection import maskrcnn_resnet50_fpn_v2, MaskRCNN_ResNet50_FPN_V2_Weights
# PyTorch TensorBoard support
from torch.utils.tensorboard import SummaryWriter
import cv2
from torchviz import make_dot
from torchvision.ops import masks_to_boxes
from tools import utils
import pandas as pd
from torch.utils.data import DataLoader, Dataset

from tools.presets import get_modules


class MaskRCNNDataset(Dataset):
    def __init__(self, dataset_path, transforms=None, dataset_type=None, enhance=False, target_type='polygon'):
        self.data_path = dataset_path
        self.transforms = transforms
        self.img_path = os.path.join(dataset_path, "images/" + dataset_type)
        self.lbl_path = os.path.join(dataset_path, "labels/" + dataset_type)
        self.imgs = os.listdir(self.img_path)
        self.lbls = os.listdir(self.lbl_path)
        self.enhance = enhance
        self.target_type = target_type
        # print('maskrcnn inited!')

    def __getitem__(self, item):
        # print('__getitem__')
        img_path = os.path.join(self.img_path, self.imgs[item])
        lbl_path = os.path.join(self.lbl_path, self.imgs[item][:-3] + 'txt')
        img = PIL.Image.open(img_path).convert('RGB')
        # h, w = np.array(img).shape[:2]
        w, h = img.size
        # print(f'h,w:{h, w}')
        target = self.read_target(item=item, lbl_path=lbl_path, shape=(h, w))
        if self.enhance:
            img = self.heatmap_enhance(img)
        if self.transforms:
            img = self.transforms(img)

        # print(f'img:{img.shape},target:{target}')
        return img, target

    def create_masks_from_polygons(self, polygons, image_shape):
        """创建一个与图像尺寸相同的掩码，并填充多边形轮廓"""
        colors = np.array([plt.cm.Spectral(each) for each in np.linspace(0, 1, len(polygons))])
        masks = []

        for polygon_data, col in zip(polygons, colors):
            mask = np.zeros(image_shape[:2], dtype=np.uint8)
            # 将多边形顶点转换为 NumPy 数组
            _, polygon = polygon_data
            pts = np.array(polygon, np.int32).reshape((-1, 1, 2))

            # 使用 OpenCV 的 fillPoly 函数填充多边形
            # print(f'color:{col[:3]}')
            cv2.fillPoly(mask, [pts], np.array(col[:3]) * 255)
            mask = torch.from_numpy(mask)
            mask[mask != 0] = 1
            masks.append(mask)

        return masks

    def read_polygons(self, lbl_path, shape):
        """读取 YOLOv8 格式的标注文件并解析多边形轮廓"""
        polygons = []
        w, h = shape[:2]
        with open(lbl_path, 'r') as f:
            lines = f.readlines()

        for line in lines:
            parts = line.strip().split()
            class_id = int(parts[0])
            # 假设多边形顶点从第2个元素开始，且已经归一化
            polygon = [float(coord) for coord in parts[1:]]

            # # 将归一化坐标转换为像素坐标
            # polygon = [int(polygon[i] * image_shape[1] if i % 2 == 0 else polygon[i] * image_shape[0]) for i in
            #            range(len(polygon))]

            points = np.array(parts[1:], dtype=np.float32).reshape(-1, 2)  # 读取点坐标
            # print(f'points :{points}')
            points[:, 0] *= h
            points[:, 1] *= w

            # 将轮廓坐标重新组织为 (x, y) 对
            # polygon = [(polygon[i], polygon[i + 1]) for i in range(0, len(polygon), 2)]
            # polygons.append((class_id, polygon))
            polygons.append((class_id, points))

        return polygons

    def read_masks_from_pixelpoints(self, lbl_path, shape):
        h, w = shape
        masks = []
        labels = []

        with open(lbl_path, 'r') as reader:
            lines = reader.readlines()
            mask_points = []
            for line in lines:
                mask = torch.zeros((h, w), dtype=torch.uint8)
                parts = line.strip().split()
                # print(f'parts:{parts}')
                cls = torch.tensor(int(parts[0]), dtype=torch.int64)
                labels.append(cls)
                x_array = parts[1::2]
                y_array = parts[2::2]

                for x, y in zip(x_array, y_array):
                    x = float(x)
                    y = float(y)
                    mask_points.append((int(y * h), int(x * w)))
                # points = np.array([[float(parts[i]), float(parts[i + 1])] for i in range(1, len(parts), 2)])
                # mask_resized = cv2.resize(points, (1600, 256), interpolation=cv2.INTER_NEAREST)
                # print(f'mask_points:{mask_points}')
                for p in mask_points:
                    mask[p] = 1
                masks.append(mask)
        reader.close()
        return labels, masks

    def read_target(self, item, lbl_path, shape):
        # print(f'lbl_path:{lbl_path}')
        h, w = shape
        labels = []
        masks = []
        if self.target_type == 'polygon':
            polygons = self.read_polygons(lbl_path, shape)
            masks = self.create_masks_from_polygons(polygons, shape)
            labels = [torch.tensor(item[0]) for item in polygons]
        elif self.target_type == 'pixel':
            labels, masks = self.read_masks_from_pixelpoints(lbl_path, shape)

        # print(f'labels:{labels}')
        target = {}
        # print(f'masks:{masks}')
        # print(f'masks:{torch.stack(masks)}')
        # print(f'boxes:{masks_to_boxes(torch.stack(masks))}')
        target["boxes"] = masks_to_boxes(torch.stack(masks))

        target["labels"] = torch.stack(labels)
        # print(f'masks: {masks[0].shape}')
        target["masks"] = torch.stack(masks)
        # target["image_id"] = image_id
        # target["area"] = area
        # target["iscrowd"] = iscrowd
        target["image_id"] = torch.tensor(item)

        target["area"] = torch.zeros(len(masks))
        target["iscrowd"] = torch.zeros(len(masks))

        # print(f'read target:{len(target)}')
        return target

    def heatmap_enhance(self, img):
        # 直方图均衡化
        img_eq = cv2.equalizeHist(img)

        # 自适应直方图均衡化
        # clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
        # img_clahe = clahe.apply(img)

        # 将灰度图转换为热力图
        heatmap = cv2.applyColorMap(img_eq, cv2.COLORMAP_HOT)

    def __len__(self):
        return len(self.imgs)


class MaskRCNNModel(nn.Module):

    def __init__(self, num_classes=3, transforms=None):
        super(MaskRCNNModel, self).__init__()
        self.__model = torchvision.models.detection.maskrcnn_resnet50_fpn_v2(
            weights=MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT)
        if transforms is None:
            self.transforms = MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT.transforms()

        self.__set_model_num_classes(num_classes)

        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    def train_one_epoch(self, optimizer, data_loader, device, epoch, print_freq, scaler=None):
        self.__model.train()
        metric_logger = utils.MetricLogger(delimiter="  ")
        metric_logger.add_meter("lr", utils.SmoothedValue(window_size=1, fmt="{value:.6f}"))
        header = f"Epoch: [{epoch}]"

        lr_scheduler = None
        if epoch == 0:
            warmup_factor = 1.0 / 1000
            warmup_iters = min(1000, len(data_loader) - 1)

            lr_scheduler = torch.optim.lr_scheduler.LinearLR(
                optimizer, start_factor=warmup_factor, total_iters=warmup_iters
            )

        for images, targets in metric_logger.log_every(data_loader, print_freq, header):
            images = list(image.to(device) for image in images)
            targets = [{k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in t.items()} for t in targets]
            # for t in targets:
            # print(f'target:{t["labels"]}')
            with torch.cuda.amp.autocast(enabled=scaler is not None):
                loss_dict = self.__model(images, targets)
                # print(f'loss_dict: {loss_dict}')  # 打印损失字典
                # print(f'loss_dict:{loss_dict}')
                losses = sum(loss for loss in loss_dict.values())

            # reduce losses over all GPUs for logging purposes
            loss_dict_reduced = utils.reduce_dict(loss_dict)
            losses_reduced = sum(loss for loss in loss_dict_reduced.values())

            loss_value = losses_reduced.item()

            if not math.isfinite(loss_value):
                print(f"Loss is {loss_value}, stopping training")
                print(loss_dict_reduced)
                sys.exit(1)

            optimizer.zero_grad()
            if scaler is not None:
                scaler.scale(losses).backward()
                scaler.step(optimizer)
                scaler.update()
            else:
                losses.backward()
                optimizer.step()

            if lr_scheduler is not None:
                lr_scheduler.step()

            metric_logger.update(loss=losses_reduced, **loss_dict_reduced)
            metric_logger.update(lr=optimizer.param_groups[0]["lr"])

        return metric_logger

    def train(self, **kwargs):
        # 默认参数
        default_params = {
            'data_dir': '/path/to/dataset',
            'num_classes': 10,
            'opt': 'adamw',
            'batch_size': 2,
            'epochs': 10,
            'lr': 0.005,
            'momentum': 0.9,
            'weight_decay': 1e-4,
            'lr_step_size': 3,
            'lr_gamma': 0.1,
            'num_workers': 4,
            'print_freq': 10,
            'target_type':'polygon',
            'enable_logs':True
        }
        # 更新默认参数
        for key, value in kwargs.items():
            if key in default_params:
                default_params[key] = value
            else:
                raise ValueError(f"Unknown argument: {key}")

        # 解析参数
        data_dir = default_params['data_dir']
        num_classes = default_params['num_classes']
        batch_size = default_params['batch_size']
        epochs = default_params['epochs']
        lr = default_params['lr']
        momentum = default_params['momentum']
        weight_decay = default_params['weight_decay']
        lr_step_size = default_params['lr_step_size']
        lr_gamma = default_params['lr_gamma']
        num_workers = default_params['num_workers']
        print_freq = default_params['print_freq']
        target_type=default_params['target_type']
        # 设置设备
        device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

        train_result_ptath = os.path.join('../train_results', datetime.now().strftime("%Y%m%d_%H%M%S"))
        wts_path = os.path.join(train_result_ptath, 'weights')
        tb_path = os.path.join(train_result_ptath, 'logs')
        writer = SummaryWriter(tb_path)

        if not os.path.exists('../train_results'):
            os.mkdir('../train_results')

        self.__set_model_num_classes(num_classes)
        self.__model.to(device)
        optimizer = torch.optim.AdamW(self.__model.parameters(), lr=lr, weight_decay=weight_decay)

        dataset = MaskRCNNDataset(dataset_path=data_dir,
                                  transforms=self.transforms, dataset_type='train',target_type=target_type)
        dataset_test = MaskRCNNDataset(dataset_path=data_dir, transforms=None,
                                       dataset_type='val')

        train_sampler = torch.utils.data.RandomSampler(dataset)
        test_sampler = torch.utils.data.SequentialSampler(dataset_test)
        train_batch_sampler = torch.utils.data.BatchSampler(train_sampler, batch_size, drop_last=True)
        train_collate_fn = utils.collate_fn
        data_loader = torch.utils.data.DataLoader(
            dataset, batch_sampler=train_batch_sampler, num_workers=num_workers, collate_fn=train_collate_fn
        )
        # data_loader_test = torch.utils.data.DataLoader(
        #     dataset_test, batch_size=1, sampler=test_sampler, num_workers=num_workers, collate_fn=utils.collate_fn
        # )



        img_results_path = os.path.join(train_result_ptath, 'img_results')
        if os.path.exists(train_result_ptath):
            pass
        #     os.remove(train_result_ptath)
        else:
            os.mkdir(train_result_ptath)

        if os.path.exists(train_result_ptath):
            os.mkdir(wts_path)
            os.mkdir(img_results_path)

        for epoch in range(epochs):
            metric_logger = self.train_one_epoch(optimizer, data_loader, device, epoch, print_freq, None)
            losses = metric_logger.meters['loss'].global_avg
            print(f'epoch {epoch}:loss:{losses}')
            if os.path.exists(f'{wts_path}/last.pt'):
                os.remove(f'{wts_path}/last.pt')
            torch.save(self.__model.state_dict(), f'{wts_path}/last.pt')

            self.write_metric_logs(epoch, metric_logger, writer)

            if epoch == 0:
                best_loss = losses;
            if best_loss >= losses:
                best_loss = losses
                if os.path.exists(f'{wts_path}/best.pt'):
                    os.remove(f'{wts_path}/best.pt')
                torch.save(self.__model.state_dict(), f'{wts_path}/best.pt')

    def write_metric_logs(self, epoch, metric_logger, writer):
        writer.add_scalar(f'loss_classifier:', metric_logger.meters['loss_classifier'].global_avg, epoch)
        writer.add_scalar(f'loss_box_reg:', metric_logger.meters['loss_box_reg'].global_avg, epoch)
        writer.add_scalar(f'loss_mask:', metric_logger.meters['loss_mask'].global_avg, epoch)
        writer.add_scalar(f'loss_objectness:', metric_logger.meters['loss_objectness'].global_avg, epoch)
        writer.add_scalar(f'loss_rpn_box_reg:', metric_logger.meters['loss_rpn_box_reg'].global_avg, epoch)
        writer.add_scalar(f'train loss:', metric_logger.meters['loss'].global_avg, epoch)

    def __set_model_num_classes(self, num_classes):
        in_features = self.__model.roi_heads.box_predictor.cls_score.in_features
        self.__model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes=num_classes)
        in_features_mask = self.__model.roi_heads.mask_predictor.conv5_mask.in_channels
        hidden_layer = 256
        self.__model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask, hidden_layer,
                                                                  num_classes=num_classes)

    def load_state_dict(self, state_dict: Mapping[str, Any], strict: bool = True):
        self.__model.load_state_dict(state_dict)
        # return super().load_state_dict(state_dict, strict)

    def predict(self, src, show_box=True, show_mask=True):
        self.__model.eval()

        img = read_image(src)
        img = self.transforms(img)
        img = img.to(self.device)
        result = self.__model([img])
        print(f'result:{result}')
        masks = result[0]['masks']
        boxes = result[0]['boxes']
        # cv2.imshow('mask',masks[0].cpu().detach().numpy())
        boxes = boxes.cpu().detach()
        drawn_boxes = draw_bounding_boxes((img * 255).to(torch.uint8), boxes, colors="red", width=5)
        print(f'drawn_boxes:{drawn_boxes.shape}')
        boxed_img = drawn_boxes.permute(1, 2, 0).numpy()
        # boxed_img=cv2.resize(boxed_img,(800,800))
        # cv2.imshow('boxes',boxed_img)

        mask = masks[0].cpu().detach().permute(1, 2, 0).numpy()

        mask = cv2.resize(mask, (800, 800))
        # cv2.imshow('mask',mask)
        img = img.cpu().detach().permute(1, 2, 0).numpy()

        masked_img = self.overlay_masks_on_image(boxed_img, masks)
        masked_img = cv2.resize(masked_img, (800, 800))
        cv2.imshow('img_masks', masked_img)
        # show_img_boxes_masks(img, boxes, masks)
        cv2.waitKey(0)

    def generate_colors(self, n):
        """
        生成n个均匀分布在HSV色彩空间中的颜色，并转换成BGR色彩空间。

        :param n: 需要的颜色数量
        :return: 一个包含n个颜色的列表，每个颜色为BGR格式的元组
        """
        hsv_colors = [(i / n * 180, 1 / 3 * 255, 2 / 3 * 255) for i in range(n)]
        bgr_colors = [tuple(map(int, cv2.cvtColor(np.uint8([[hsv]]), cv2.COLOR_HSV2BGR)[0][0])) for hsv in hsv_colors]
        return bgr_colors

    def overlay_masks_on_image(self, image, masks, alpha=0.6):
        """
        在原图上叠加多个掩码，每个掩码使用不同的颜色。

        :param image: 原图 (NumPy 数组)
        :param masks: 掩码列表 (每个都是 NumPy 数组，二值图像)
        :param colors: 颜色列表 (每个颜色都是 (B, G, R) 格式的元组)
        :param alpha: 掩码的透明度 (0.0 到 1.0)
        :return: 叠加了多个掩码的图像
        """
        colors = self.generate_colors(len(masks))
        if len(masks) != len(colors):
            raise ValueError("The number of masks and colors must be the same.")

        # 复制原图，避免修改原始图像
        overlay = image.copy()

        for mask, color in zip(masks, colors):
            # 确保掩码是二值图像
            mask = mask.cpu().detach().permute(1, 2, 0).numpy()
            binary_mask = (mask > 0).astype(np.uint8) * 255  # 你可以根据实际情况调整阈值

            # 创建彩色掩码
            colored_mask = np.zeros_like(image)
            colored_mask[:] = color
            colored_mask = cv2.bitwise_and(colored_mask, colored_mask, mask=binary_mask)

            # 将彩色掩码与当前的叠加图像混合
            overlay = cv2.addWeighted(overlay, 1 - alpha, colored_mask, alpha, 0)

        return overlay


if __name__ == '__main__':
    ins_model = MaskRCNNModel(num_classes=5);
    data_path = r'F:\DevTools\datasets\renyaun\1012\spilt'
    # ins_model.train(data_dir=data_path,epochs=5000,target_type='pixel',batch_size=6,num_workers=10,num_classes=5)
    ins_model.train(data_dir=data_path, epochs=5000,  batch_size=2, num_workers=10, num_classes=5)
    # img_path = r"F:\DevTools\datasets\renyaun\1012\spilt\images\train\2024-09-23-10-13-26_SaveImage.png"
    # state_dict = torch.load(r"F:\Downloads\maskrcnn_wts\renyuan.pt")
    # ins_model.load_state_dict(state_dict)
    # ins_model.predict(img_path)
