import os
import cv2
import numpy as np
import matplotlib.pyplot as plt
import torch


def read_label_file(label_path, image_shape):
    """读取 YOLOv8 格式的标注文件并解析多边形轮廓"""
    polygons = []
    w,h=image_shape[:2]
    with open(label_path, 'r') as f:
        lines = f.readlines()

    for line in lines:
        parts = line.strip().split()
        class_id = int(parts[0])
        # 假设多边形顶点从第2个元素开始，且已经归一化
        # polygon = [float(coord) for coord in parts[1:]]

        # # 将归一化坐标转换为像素坐标
        # polygon = [int(polygon[i] * image_shape[1] if i % 2 == 0 else polygon[i] * image_shape[0]) for i in
        #            range(len(polygon))]

        points = np.array(parts[1:], dtype=np.float32).reshape(-1, 2)  # 读取点坐标
        # print(f'points :{points}')
        # points[:, 0] *= h
        # points[:, 1] *= w


        # 将轮廓坐标重新组织为 (x, y) 对
        # polygon = [(polygon[i], polygon[i + 1]) for i in range(0, len(polygon), 2)]
        # polygons.append((class_id, polygon))
        polygons.append((class_id, points))
        # polygons.append((cl, points))

    return polygons

def read_masks_from_pixelpoints(lbl_path, shape):
    h, w = shape
    masks = []
    labels = []

    with open(lbl_path, 'r') as reader:
        lines = reader.readlines()
        mask_points = []
        for line in lines:
            mask = torch.zeros((h, w), dtype=torch.uint8)
            parts = line.strip().split()
            # print(f'parts:{parts}')
            cls = torch.tensor(int(parts[0]), dtype=torch.int64)
            labels.append(cls)
            x_array = parts[1::2]
            y_array = parts[2::2]

            for x, y in zip(x_array, y_array):
                x = float(x)
                y = float(y)
                mask_points.append((int(y * h), int(x * w)))
            # points = np.array([[float(parts[i]), float(parts[i + 1])] for i in range(1, len(parts), 2)])
            # mask_resized = cv2.resize(points, (1600, 256), interpolation=cv2.INTER_NEAREST)
            # print(f'mask_points:{mask_points}')
            for p in mask_points:
                mask[p] = 1
            masks.append(mask)
    reader.close()
    return labels, masks

def create_mask_from_polygons(polygons, image_shape):
    """创建一个与图像尺寸相同的掩码，并填充多边形轮廓"""
    colors = np.array([plt.cm.Spectral(each) for each in np.linspace(0, 1, len(polygons))])
    masks=[]


    for polygon_data ,col in zip(polygons,colors):
        mask = np.zeros(image_shape[:2], dtype=np.uint8)
        # 将多边形顶点转换为 NumPy 数组
        _,polygon=polygon_data
        pts = np.array(polygon, np.int32).reshape((-1, 1, 2))

        # 使用 OpenCV 的 fillPoly 函数填充多边形
        print(f'color:{col[:3]}')
        cv2.fillPoly(mask, [pts], np.array(col[:3]) * 255)
        print(f'mask: {mask>0}')
        masks.append(mask)

    return masks


def display_image_with_mask(image, masks, alpha=0.5):
    """将掩码叠加到图像上并显示"""
    # # 将掩码扩展为三通道
    # mask_3ch = cv2.cvtColor(mask, cv2.COLOR_GRAY2BGR)
    #
    # # 将掩码叠加到图像上
    # overlay = cv2.addWeighted(image, 1, mask_3ch, alpha, 0)
    #
    # # 显示结果
    # plt.imshow(cv2.cvtColor(overlay, cv2.COLOR_BGR2RGB))
    # plt.axis('off')
    # plt.show()
    fig, axs = plt.subplots(2, 1)
    # print(f'test_img:{test_img * 255}')
    axs[0].imshow(image)
    axs[0].axis('off')
    axs[1].axis('off')
    # axs[1].imshow(image * 255)
    for img_mask in masks:
        # img_mask=img_mask.unsqueeze(0)
        # img_mask = img_mask.expand_as(test_img)
        # print(f'img_mask:{img_mask.shape}')
        axs[1].imshow(img_mask, alpha=1)
    plt.show()

# 示例
image_path = r"I:\wireframe_yolo\images\train\00030210_3.png"
label_path =  r"I:\wireframe_yolo\labels\train\00030210_3.txt"

# 读取图像
image = cv2.imread(image_path)
print(f'image shape:{image.shape}')
image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

# 读取并解析多边形轮廓
polygons = read_label_file(label_path, image.shape)

labels,mask=read_masks_from_pixelpoints(label_path, image.shape[:2])
# 创建掩码
# mask = create_mask_from_polygons(polygons, image.shape)
# print(f'mask:{mask}')

# 显示带有掩码的图像
display_image_with_mask(image, mask, alpha=1)