import math
import os
import sys
from datetime import datetime

import torch
import torchvision
from torchvision.models.detection import MaskRCNN_ResNet50_FPN_V2_Weights
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor

from MaskRCNN import MaskRCNNDataset
from tools import utils
from tools.engine import evaluate

device = (
    "cuda"
    if torch.cuda.is_available()
    else "mps"
    if torch.backends.mps.is_available()
    else "cpu"
)
def train_one_epoch(model, optimizer, data_loader, device, epoch, print_freq, scaler=None):
    model.train()
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter("lr", utils.SmoothedValue(window_size=1, fmt="{value:.6f}"))
    header = f"Epoch: [{epoch}]"

    lr_scheduler = None
    if epoch == 0:
        warmup_factor = 1.0 / 1000
        warmup_iters = min(1000, len(data_loader) - 1)

        lr_scheduler = torch.optim.lr_scheduler.LinearLR(
            optimizer, start_factor=warmup_factor, total_iters=warmup_iters
        )

    for images, targets in metric_logger.log_every(data_loader, print_freq, header):
        images = list(image.to(device) for image in images)
        targets = [{k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in t.items()} for t in targets]
        # for t in targets:
            # print(f'target:{t["labels"]}')
        with torch.cuda.amp.autocast(enabled=scaler is not None):
            loss_dict = model(images, targets)
            # print(f'loss_dict: {loss_dict}')  # 打印损失字典
            # print(f'loss_dict:{loss_dict}')
            losses = sum(loss for loss in loss_dict.values())

        # reduce losses over all GPUs for logging purposes
        loss_dict_reduced = utils.reduce_dict(loss_dict)
        losses_reduced = sum(loss for loss in loss_dict_reduced.values())

        loss_value = losses_reduced.item()

        if not math.isfinite(loss_value):
            print(f"Loss is {loss_value}, stopping training")
            print(loss_dict_reduced)
            sys.exit(1)

        optimizer.zero_grad()
        if scaler is not None:
            scaler.scale(losses).backward()
            scaler.step(optimizer)
            scaler.update()
        else:
            losses.backward()
            optimizer.step()

        if lr_scheduler is not None:
            lr_scheduler.step()





        metric_logger.update(loss=losses_reduced, **loss_dict_reduced)
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])

    return metric_logger
if __name__ == '__main__':
    num_classes = 5
    # models = torchvision.models.get_model(
    #     'maskrcnn_resnet50_fpn_v2', weights=MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT, num_classes=num_classes
    # )
    model=torchvision.models.detection.maskrcnn_resnet50_fpn_v2(weights=MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT)
    in_features = model.roi_heads.box_predictor.cls_score.in_features

    model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes=5)
    in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
    hidden_layer = 256
    model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask, hidden_layer, num_classes=5)
    transforms=MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT.transforms()
    print(f'transforms:{transforms}')
    model.to(device)
    # dataset=MaskRCNNDataset(dataset_path=r'/data/rlq/datasets/severstal-steel-defect-detection',transforms=transforms,dataset_type='train')
    # dataset_test = MaskRCNNDataset(dataset_path=r'/data/rlq/datasets/severstal-steel-defect-detection', transforms=None, dataset_type='val')

    datasetpath=r'F:\DevTools\datasets\renyaun\1012\spilt'

    dataset = MaskRCNNDataset(dataset_path=datasetpath,
                              transforms=transforms, dataset_type='train')
    dataset_test = MaskRCNNDataset(dataset_path=datasetpath, transforms=None,
                                   dataset_type='val')

    train_sampler = torch.utils.data.RandomSampler(dataset)
    test_sampler = torch.utils.data.SequentialSampler(dataset_test)
    batch_size=2
    train_batch_sampler = torch.utils.data.BatchSampler(train_sampler, batch_size, drop_last=True)
    train_collate_fn = utils.collate_fn
    data_loader = torch.utils.data.DataLoader(
        dataset, batch_sampler=train_batch_sampler, num_workers=10, collate_fn=train_collate_fn
    )
    data_loader_test = torch.utils.data.DataLoader(
        dataset_test, batch_size=1, sampler=test_sampler, num_workers=10, collate_fn=utils.collate_fn
    )
    epochs=5000
    lr=0.001
    weight_decay=0.0005
    print_freq=10

    optimizer = torch.optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay)

    wts_dir_ptath = os.path.join('weights', datetime.now().strftime("%Y%m%d_%H%M%S"))
    if os.path.exists(wts_dir_ptath):
        os.remove(wts_dir_ptath)
    os.mkdir(wts_dir_ptath)
    best_loss = 10000

    for epoch in range(epochs):
       metric_logger= train_one_epoch(model, optimizer, data_loader, device, epoch, print_freq, None)
       losses=metric_logger.meters['loss'].global_avg
       print(f'epoch {epoch}:loss:{losses}')

       if os.path.exists(f'{wts_dir_ptath}/last.pt'):
           os.remove(f'{wts_dir_ptath}/last.pt')
       torch.save(model.state_dict(), f'{wts_dir_ptath}/last.pt')

       if epoch == 0:
           best_loss = losses;
       if best_loss >= losses:
           best_loss = losses
           if os.path.exists(f'{wts_dir_ptath}/best.pt'):
               os.remove(f'{wts_dir_ptath}/best.pt')
           torch.save(model.state_dict(), f'{wts_dir_ptath}/best.pt')
        # evaluate(models, data_loader, device)